@@ -40,7 +40,7 @@
 #include <mipict.h>
 
 #if 0
-#define __DBG(x) ErrorF x
+#define __DBG(x) LogF x
 #else
 #define __DBG(x)
 #endif
@@ -390,17 +390,17 @@ mono_merge_unsorted_edges(struct mono_edge *head, struct mono_edge *unsorted)
 static inline void
 __dbg_mono_edges(const char *function, struct mono_edge *edges)
 {
-	ErrorF("%s: ", function);
+	DBG(("%s: ", function));
 	while (edges) {
 		if (edges->x.quo < INT16_MAX << 16) {
-			ErrorF("(%d.%06d)+(%d.%06d)x%d, ",
-			       edges->x.quo, edges->x.rem,
-			       edges->dxdy.quo, edges->dxdy.rem,
-			       edges->dy*edges->dir);
+			DBG(("(%d.%06d)+(%d.%06d)x%d, ",
+			     edges->x.quo, edges->x.rem,
+			     edges->dxdy.quo, edges->dxdy.rem,
+			     edges->dy*edges->dir));
 		}
 		edges = edges->next;
 	}
-	ErrorF("\n");
+	DBG(("\n"));
 }
 #define DBG_MONO_EDGES(x) __dbg_mono_edges(__FUNCTION__, x)
 static inline void
@@ -813,7 +813,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 				       mono.clip.extents.x1,  mono.clip.extents.y1,
 				       mono.clip.extents.x2 - mono.clip.extents.x1,
 				       mono.clip.extents.y2 - mono.clip.extents.y1,
-				       memset(&mono.op, 0, sizeof(mono.op))))
+				       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op))))
 		return false;
 
 	num_threads = 1;
@@ -852,7 +852,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 			threads[n].extents.y1 = y;
 			threads[n].extents.y2 = y += h;
 
-			sna_threads_run(mono_span_thread, &threads[n]);
+			sna_threads_run(n, mono_span_thread, &threads[n]);
 		}
 
 		threads[0].extents.y1 = y;
@@ -924,7 +924,6 @@ mono_trapezoids_span_converter(struct sna *sna,
 				      traps[n].top, traps[n].bottom,
 				      &traps[n].right.p1, &traps[n].right.p2, -1);
 		}
-		memset(&mono.op, 0, sizeof(mono.op));
 		if (mono.sna->render.composite(mono.sna,
 					       PictOpClear,
 					       mono.sna->clear, NULL, dst,
@@ -933,7 +932,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 					       mono.clip.extents.x1,  mono.clip.extents.y1,
 					       mono.clip.extents.x2 - mono.clip.extents.x1,
 					       mono.clip.extents.y2 - mono.clip.extents.y1,
-					       &mono.op)) {
+					       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 			mono_render(&mono);
 			mono.op.done(mono.sna, &mono.op);
 		}
@@ -1169,7 +1168,10 @@ unbounded_pass:
 		mono.span = mono_span__fast;
 	else
 		mono.span = mono_span;
-	mono_render(&mono);
+	if (sigtrap_get() == 0) {
+		mono_render(&mono);
+		sigtrap_put();
+	}
 	mono_fini(&mono);
 
 	if (op) {
@@ -1273,14 +1275,13 @@ mono_trap_span_converter(struct sna *sna,
 			      &p1, &p2, -1);
 	}
 
-	memset(&mono.op, 0, sizeof(mono.op));
 	if (mono.sna->render.composite(mono.sna, PictOpAdd, src, NULL, dst,
-					0, 0,
-					0, 0,
-					mono.clip.extents.x1,  mono.clip.extents.y1,
-					mono.clip.extents.x2 - mono.clip.extents.x1,
-					mono.clip.extents.y2 - mono.clip.extents.y1,
-					&mono.op)) {
+				       0, 0,
+				       0, 0,
+				       mono.clip.extents.x1,  mono.clip.extents.y1,
+				       mono.clip.extents.x2 - mono.clip.extents.x1,
+				       mono.clip.extents.y2 - mono.clip.extents.y1,
+				       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 		mono_render(&mono);
 		mono.op.done(mono.sna, &mono.op);
 	}
@@ -1356,7 +1357,6 @@ mono_triangles_span_converter(struct sna *sna,
 			      &tri[n].p3, &tri[n].p1, 1);
 	}
 
-	memset(&mono.op, 0, sizeof(mono.op));
 	if (mono.sna->render.composite(mono.sna, op, src, NULL, dst,
 				       src_x + mono.clip.extents.x1 - dst_x - dx,
 				       src_y + mono.clip.extents.y1 - dst_y - dy,
@@ -1364,7 +1364,7 @@ mono_triangles_span_converter(struct sna *sna,
 				       mono.clip.extents.x1,  mono.clip.extents.y1,
 				       mono.clip.extents.x2 - mono.clip.extents.x1,
 				       mono.clip.extents.y2 - mono.clip.extents.y1,
-				       &mono.op)) {
+				       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 		if (mono.clip.data == NULL && mono.op.damage == NULL)
 			mono.span = mono_span__fast;
 		else
@@ -1402,7 +1402,6 @@ mono_triangles_span_converter(struct sna *sna,
 				      &tri[n].p3, &tri[n].p1, 1);
 		}
 
-		memset(&mono.op, 0, sizeof(mono.op));
 		if (mono.sna->render.composite(mono.sna,
 					       PictOpClear,
 					       mono.sna->clear, NULL, dst,
@@ -1411,7 +1410,7 @@ mono_triangles_span_converter(struct sna *sna,
 					       mono.clip.extents.x1,  mono.clip.extents.y1,
 					       mono.clip.extents.x2 - mono.clip.extents.x1,
 					       mono.clip.extents.y2 - mono.clip.extents.y1,
-					       &mono.op)) {
+					       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 			if (mono.clip.data == NULL && mono.op.damage == NULL)
 				mono.span = mono_span__fast;
 			else
