@@ -205,10 +205,10 @@ sna_video_sprite_show(struct sna *sna,
 	/* XXX handle video spanning multiple CRTC */
 
 	VG_CLEAR(s);
-	s.plane_id = sna_crtc_to_plane(crtc);
+	s.plane_id = sna_crtc_to_sprite(crtc);
 
 	update_dst_box_to_crtc_coords(sna, crtc, dstBox);
-	if (crtc->rotation & (RR_Rotate_90 | RR_Rotate_270)) {
+	if (video->rotation & (RR_Rotate_90 | RR_Rotate_270)) {
 		int tmp = frame->width;
 		frame->width = frame->height;
 		frame->height = tmp;
@@ -242,6 +242,7 @@ sna_video_sprite_show(struct sna *sna,
 	if (frame->bo->delta == 0) {
 		uint32_t offsets[4], pitches[4], handles[4];
 		uint32_t pixel_format;
+		bool purged = true;
 
 		handles[0] = frame->bo->handle;
 		pitches[0] = frame->pitch[0];
@@ -250,9 +251,11 @@ sna_video_sprite_show(struct sna *sna,
 		switch (frame->id) {
 		case FOURCC_RGB565:
 			pixel_format = DRM_FORMAT_RGB565;
+			purged = sna->scrn->depth != 16;
 			break;
 		case FOURCC_RGB888:
 			pixel_format = DRM_FORMAT_XRGB8888;
+			purged = sna->scrn->depth != 24;
 			break;
 		case FOURCC_UYVY:
 			pixel_format = DRM_FORMAT_UYVY;
@@ -277,6 +280,8 @@ sna_video_sprite_show(struct sna *sna,
 		}
 
 		frame->bo->scanout = true;
+		/* Don't allow the scanout to be cached if not suitable for front */
+		frame->bo->purged = purged;
 	}
 
 	assert(frame->bo->scanout);
@@ -310,7 +315,10 @@ sna_video_sprite_show(struct sna *sna,
 		if (video->plane) {
 			memset(&s, 0, sizeof(s));
 			s.plane_id = video->plane;
-			drmIoctl(video->sna->kgem.fd, DRM_IOCTL_MODE_SETPLANE, &s);
+			if (drmIoctl(video->sna->kgem.fd, DRM_IOCTL_MODE_SETPLANE, &s)) {
+				DBG(("SET_PLANE failed to turn off existing sprite: ret=%d\n", errno));
+				return false;
+			}
 		}
 		video->plane = s.plane_id;
 	}
@@ -375,11 +383,15 @@ static int sna_video_sprite_put_image(ClientPtr client,
 				   &clip))
 		goto invisible;
 
-	if (!crtc || sna_crtc_to_plane(crtc) == 0)
+	if (!crtc || sna_crtc_to_sprite(crtc) == 0)
 		goto invisible;
 
-	/* sprites can't handle rotation natively, store it for the copy func */
-	video->rotation = crtc->rotation;
+	/* if sprite can't handle rotation natively, store it for the copy func */
+	video->rotation = RR_Rotate_0;
+	if (!sna_crtc_set_sprite_rotation(crtc, crtc->rotation)) {
+		sna_crtc_set_sprite_rotation(crtc, RR_Rotate_0);
+		video->rotation = crtc->rotation;
+	}
 
 	if (xvmc_passthrough(format->id)) {
 		DBG(("%s: using passthough, name=%d\n",
